/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>
#include <errno.h>

#include <gnome.h>

#include "articlelist.h"
#include "dialogs/dialogs.h"
#include "fnmatch.h"
#include "globals.h"
#include "grouplist.h"
#include "gui.h"
#include "pan-db.h"
#include "prefs.h"
#include "queue.h"
#include "task-headers.h"
#include "task-grouplist.h"
#include "util.h"

#include "xpm/folder.xpm"
#include "xpm/envelope.xpm"


static void grouplist_update_ui_thread (void* data);

static void group_dialog_clicked_cb (GnomeDialog*, int, Server*);
static void grouplist_add_group (const Server*, Group*);
static void grouplist_update_selected_rows (void);
static void grouplist_grep (GtkWidget *group_entry);

extern GtkWidget* ttips;

static GnomeUIInfo groups_download_menu[] =
{
	{
		GNOME_APP_UI_ITEM,
		N_("Download..."),
		N_("Download..."),
		grouplist_selected_download_dialog
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("Get New Article Headers"),
		N_("Get New Article Headers"),
		grouplist_selected_download_new
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Sample Article Headers"),
		N_("Sample Article Headers"),
		grouplist_selected_download_sample
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Get All Article Headers"),
		N_("Get All Article Headers"),
		grouplist_selected_download_all
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("Get New Article Headers and Bodies"),
		N_("Get New Article Headers and Bodies"),
		grouplist_selected_download_new_bodies
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Sample Article Headers and Bodies"),
		N_("Sample Article Headers and Bodies"),
		grouplist_selected_download_sample_bodies
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Get All Article Headers and Bodies"),
		N_("Get All Article Headers and Bodies"),
		grouplist_selected_download_all_bodies
	},
	GNOMEUIINFO_END
};


static GnomeUIInfo groups_menu_popup[] =
{
	{
		GNOME_APP_UI_SUBTREE,
		N_("Download Articles"),
		N_("Download Articles"),
		&groups_download_menu, NULL,
		NULL,
		GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_BLANK,
		'\0', 0, NULL
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("Refresh Article Count"),
		N_("Refresh Article Count"),
		grouplist_selected_update_count_info
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Catch Up"),
		N_("Catch Up"),
		grouplist_selected_catchup
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("Subscribe"),
		N_("Subscribe"),
		grouplist_selected_subscribe, NULL, NULL,
		GNOME_APP_PIXMAP_DATA, envelope_xpm,
		0, 0, NULL
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Unsubscribe"),
		N_("Unsubscribe"),
		grouplist_selected_unsubscribe
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("Empty Group"),
		N_("Remove all articles from the selected groups."),
		grouplist_selected_empty
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("Remove `New' Status"),
		N_("Remove selected group(s) from `new groups' list."),
		grouplist_selected_unnew
	},
	GNOMEUIINFO_SEPARATOR,
	{
		GNOME_APP_UI_ITEM,
		N_("Properties..."),
		N_("Properties..."),
		grouplist_selected_properties, NULL, NULL,
		GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PROP,
		0, 0, NULL
	},
	GNOMEUIINFO_END
};

static GnomePixmap * folder_pixmap        = NULL;
static GnomePixmap * group_pixmap         = NULL;

static int           group_mode           = GROUP_ALL;
static GtkWidget   * group_clist_menu     = NULL;
static Server      * my_server            = NULL;
static GtkWidget   * grouplist_mode_menu  = NULL;


GSList*
grouplist_get_selected_groups (void)
{
	GSList *groups = NULL;
	GList *l = NULL;
	for (l=GTK_CLIST(Pan.group_clist)->selection; l; l=l->next)
	{
		Group *group = (Group *)
			gtk_clist_get_row_data (
				GTK_CLIST (Pan.group_clist),
				GPOINTER_TO_INT (l->data));
		groups = g_slist_prepend (groups, group);
	}
	groups = g_slist_reverse (groups);
	return groups;
	
}

Server *
grouplist_get_current_server (void)
{
	return (my_server);
}

int
grouplist_get_view_mode (void)
{
	return (group_mode);
}

static void
group_dialog_clicked_cb (GnomeDialog* dialog, int num, Server *server)
{
	switch (num) {
		case 0: /* OK button */
			queue_add (TASK(task_grouplist_new (
						server, TRUE, GROUPLIST_ALL)));
		default:
			gtk_widget_destroy (GTK_WIDGET(dialog));
			break;
	}
}                                                                               

void
grouplist_set_server (Server * server)
{
	gchar * pch;
	gboolean is_file;

	g_return_if_fail (server != NULL);

	/* ??
	if (!folders->len)
		folderlist_load (); */

	/* nothing to do */
	if (my_server == server)
		return;

	/* Clear the ui and set the active list */
	gtk_clist_clear (GTK_CLIST(Pan.group_clist));
	my_server = server;

	/* already loaded the groups */
	if (server->groups->len) {
		grouplist_update_ui (server);
		return;
	}

	/* haven't loaded the groups, but found a file for it */
	pch = g_strdup_printf ("/%s/%s_grouplist.db", data_dir, server->name);
	is_file = g_file_test (pch, G_FILE_TEST_ISFILE);
	g_free (pch);
	if (is_file) {
		grouplist_load (server);
		return;
	}

	/* can't even find a file for the groups;
	   shall we download a group list? */
	{
		GtkWidget * d, * l;
		gchar * s;
		d = gnome_dialog_new (_("No Group List"),
	                              GNOME_STOCK_BUTTON_YES,
	                              GNOME_STOCK_BUTTON_NO,
	                              NULL);
		s = g_strdup_printf (_("We don't have a group list for server "
		                       "\"%s\"\nShall we get one?\n"),
		                       server->name);
		l = gtk_label_new (s);
		g_free (s);
		gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(d)->vbox), l);
		gtk_signal_connect (GTK_OBJECT(d), "clicked",
		                    GTK_SIGNAL_FUNC (group_dialog_clicked_cb),
				    server);
		gui_popup_draw (GTK_WIDGET(d), Pan.window);
	}
}

/* process what happens when a user lcicks on a group in the group list */
static void
grouplist_select_row (GtkWidget *widget, int row, int col, GdkEvent *event)
{
	GdkEventButton *bevent = (GdkEventButton *)event;
	if (!bevent)
	       return;

	if (bevent->button==1 && bevent->type==GDK_2BUTTON_PRESS)
	{
		/* which group? */
		Group* group = (Group *) gtk_clist_get_row_data(GTK_CLIST(widget), row);

		/* turn the page if in notebook layout. */
		gui_page_set (ARTICLELIST_PAGE, Pan.article_ctree);

		/* load the specified group */
		if (articlelist_get_current_group() != group)
			articlelist_set_current_group (group);
	}
}

static void
grouplist_process_selected (int type)
{
	GtkCList * clist = GTK_CLIST(Pan.group_clist);
	const gboolean has_selection = clist->selection != NULL;
	Group * group = NULL;
	GList * list = NULL;
	GList * l = NULL;

	/* build the list of groups */
	if (has_selection) {
		list = g_list_copy (GTK_CLIST (Pan.group_clist)->selection);
		for (l=list; l!=NULL; l=l->next)
			l->data = gtk_clist_get_row_data (
				GTK_CLIST (Pan.group_clist),
				GPOINTER_TO_INT (l->data));
	}

	if (!list && (group=articlelist_get_current_group()))
	{
		switch (type)
		{
			case 3:
				queue_add (TASK(task_headers_new(
					group, HEADERS_ALL, TRUE)));
				break;
			case 4:
				queue_add (TASK(task_headers_new(
					group, HEADERS_NEW, TRUE)));
				break;
			default:
				break;
		}
	}

	for (l=list; l!=NULL; l=l->next)
	{
		group = (Group*) l->data;

		switch (type)
		{
			case 1: /* subscribe */
				group_set_subscribed (group, TRUE);
				break;

			case 2: /* unsubscribe */
				group_set_subscribed (group, FALSE);
				break;
				break;

			case 3: /* download all */
				queue_add (TASK(task_headers_new(
					group, HEADERS_ALL, FALSE)));
				break;

			case 4: /* download new */
				queue_add (TASK(task_headers_new(
					group, HEADERS_NEW, FALSE)));
				break;

			case 5: /* clean/empty */
				group_empty_dialog (group);
				break;

			case 6: /* group properties */
				dialog_group_props (group);
				break;

			case 7: /* remove "new" mark */
				group_set_new (group, FALSE);
				break;

			case 8: /* remove group */
				group_remove_dialog (group);
				break;

#if 0
			case 9: /* rename */
				rename_folder_dialog (group);
				break;
#endif

			case 10: /* download all bodies */
				group_download_all_articles (group);
				break;

			default:
				break;
		}
	}

	/* if unsubscribe and in subscribe mode, remove the groups */
	if (type==2 && group_mode==GROUP_SUBSCRIBED)
		for (l=list; l!=NULL; l=l->next)
			grouplist_remove_row ((const Group*)l->data);

	/* refresh the modified entries */
	grouplist_update_selected_rows ();

	/* unselect */
	gtk_clist_unselect_all (GTK_CLIST(Pan.group_clist));

	/* cleanup */
	g_list_free (list);
}

static GPtrArray*
grouplist_get_target_groups (void)
{
	GtkCList * clist = GTK_CLIST(Pan.group_clist);
	GPtrArray * retval = NULL;
	GList * l = NULL;

	/* initialize the gptrarray */
       	retval = g_ptr_array_new ();
	pan_g_ptr_array_reserve (retval, g_list_length(clist->selection));

	/* get the group selected items; otherwise, 
	   use the articlelist's group */
	for (l=clist->selection; l!=NULL; l=l->next)
		g_ptr_array_add (retval, gtk_clist_get_row_data (clist, GPOINTER_TO_INT (l->data)));
	if (!retval->len) {
		Group * group = articlelist_get_current_group ();
		if (group != NULL)
			g_ptr_array_add (retval, group);
	}

	return retval;
}

void
grouplist_selected_subscribe (void)
{
	grouplist_process_selected (1);
}
void
grouplist_selected_unsubscribe (void)
{
	grouplist_process_selected (2);
}
void
grouplist_selected_empty (void)
{
	grouplist_process_selected (5);
}
void
grouplist_selected_properties (void)
{
	grouplist_process_selected (6);
}
void
grouplist_selected_unnew (void)
{
	grouplist_process_selected (7);
}
void
grouplist_selected_remove (void)
{
	grouplist_process_selected (8);
}

void
grouplist_selected_update_count_info (void)
{
	GPtrArray * a = grouplist_get_target_groups ();
	if (a != NULL)
		group_update_count_info (a);
}

void
grouplist_selected_catchup (void)
{
	GPtrArray * a = grouplist_get_target_groups ();
	if (a != NULL)
		pan_g_ptr_array_foreach (a, (GFunc)group_catchup, NULL);
	g_ptr_array_free (a, TRUE);
}


/**
***  Process all the commands to download
***  [new | all | sample] [articles | headers]
**/

static void
grouplist_selected_task_headers (HeaderDownloadType type,
                                 gboolean bodies)
{
	GPtrArray * a = grouplist_get_target_groups ();
	int i;

	/* sample all the selected groups */
	for (i=0; i<a->len; ++i)
	{
		Group * group = (Group*) g_ptr_array_index (a, i);
		PanObject * o = bodies
			? task_headers_new_with_bodies (group, type)
			: task_headers_new (group, type, FALSE);
		queue_add (TASK(o));
	}

	/* cleanup */
	grouplist_update_selected_rows ();
	gtk_clist_unselect_all (GTK_CLIST(Pan.group_clist));
	g_ptr_array_free (a, TRUE);
}

void
grouplist_selected_download_dialog (void)
{
	int i;
	GPtrArray * a = grouplist_get_target_groups ();

	/* sample all the selected groups */
	for (i=0; i<a->len; ++i)
		dialog_download_headers (GROUP(g_ptr_array_index(a,i)));

	/* cleanup */
	grouplist_update_selected_rows ();
	gtk_clist_unselect_all (GTK_CLIST(Pan.group_clist));
	g_ptr_array_free (a, TRUE);
}

void
grouplist_selected_download_sample (void)
{
	grouplist_selected_task_headers (HEADERS_SAMPLE, FALSE);
}
void
grouplist_selected_download_sample_bodies (void)
{
	grouplist_selected_task_headers (HEADERS_SAMPLE, TRUE);
}
void
grouplist_selected_download_all (void)
{
	grouplist_selected_task_headers (HEADERS_ALL, FALSE);
}
void
grouplist_selected_download_all_bodies (void)
{
	grouplist_selected_task_headers (HEADERS_ALL, TRUE);
}
void
grouplist_selected_download_new (void)
{
	grouplist_selected_task_headers (HEADERS_NEW, FALSE);
}
void
grouplist_selected_download_new_bodies (void)
{
	grouplist_selected_task_headers (HEADERS_NEW, TRUE);
}




#if 0
void grouplist_selected_rename (void) { grouplist_process_selected (9); }
#endif

void
grouplist_get_all (void)
{
	if (my_server == NULL)
		pan_error_dialog ("No server selected.");
	else
		queue_add (
			TASK(task_grouplist_new (
					my_server, TRUE, GROUPLIST_ALL)));
}

void
grouplist_get_new (void)
{
	if (my_server == NULL)
		pan_error_dialog ("No server selected.");
	else
		queue_add (
			TASK(task_grouplist_new (
					my_server, TRUE, GROUPLIST_NEW)));
}

static void
grouplist_all_subscribed_gfunc (gpointer data, gpointer user_data)
{
	HeaderDownloadType dl_type;
	Group *group;

	g_return_if_fail (user_data != NULL);
	g_return_if_fail (data != NULL);

	dl_type = GPOINTER_TO_INT (user_data);
	group = data;

	if (group_is_subscribed (group))
		queue_add (TASK(task_headers_new(group, dl_type, FALSE)));
}

void
grouplist_all_subscribed_download_new (void)
{
	if (my_server == NULL)
		pan_error_dialog ("No server selected.");
	else
	{
		int i;
		for (i=0; i<my_server->groups->len; ++i)
			grouplist_all_subscribed_gfunc (
				g_ptr_array_index(my_server->groups,i),
				GINT_TO_POINTER(HEADERS_NEW));
	}
}

void
grouplist_all_subscribed_download_all (void)
{
	if (my_server == NULL)
		pan_error_dialog ("No server selected.");
	else
	{
		int i;
		for (i=0; i<my_server->groups->len; ++i)
			grouplist_all_subscribed_gfunc (
				g_ptr_array_index(my_server->groups,i),
				GINT_TO_POINTER(HEADERS_ALL));
	}
}

void
grouplist_select_all (void)
{
	pan_lock();
	gtk_clist_select_all (GTK_CLIST (Pan.group_clist));
	pan_unlock();
}


static void
grouplist_button_press (GtkWidget *widget, GdkEventButton *bevent)
{
	if (bevent->button != 1)
		gtk_signal_emit_stop_by_name (
			GTK_OBJECT(widget), "button_press_event");
	
	if (bevent->button == 3)
	{
#if 0
		const int sel = GTK_CLIST(Pan.group_clist)->selection != NULL;
		GList* l = GTK_MENU_SHELL(group_clist_menu)->children; /* sub */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);
		l = l->next; /* l is unsub */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);

		l = l->next; /* l is separator */
		l = l->next; /* l is "get all headers" */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);
		l = l->next; /* l is "get new headers" */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);
		l = l->next; /* l is "sample article headers" */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);

		l = l->next; /* l is separator */
		l = l->next; /* l is "get all headers + bodies" */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);
		l = l->next; /* l is "get new headers + bodies" */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);
		l = l->next; /* l is "sample article headers + bodies" */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);

		l = l->next; /* l is separator */
		l = l->next; /* l is "empty group" */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);

		l = l->next; /* l is separator */
		l = l->next; /* l is "unnew" */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);
		l = l->next; /* l is "catchup" */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);

		l = l->next; /* l is separator */
		l = l->next; /* l is "properties" */
		gtk_widget_set_sensitive (GTK_WIDGET(l->data), sel);
#endif

		gtk_menu_popup (GTK_MENU(group_clist_menu), NULL, NULL, NULL,
				NULL, bevent->button, bevent->time);
	}
}

void
grouplist_set_view_mode (GtkWidget *widget, gpointer data)
{
	GtkWidget *menu = NULL;

	group_mode = GPOINTER_TO_INT (data);
	grouplist_update_ui (grouplist_get_current_server());

        pan_lock ();
        menu = gtk_option_menu_get_menu (GTK_OPTION_MENU(grouplist_mode_menu));
        gtk_object_ref (GTK_OBJECT(menu));
        gtk_option_menu_remove_menu (GTK_OPTION_MENU (grouplist_mode_menu));
	switch (group_mode) {
		case GROUP_ALL:
			gtk_menu_set_active(GTK_MENU (menu), 0);
			break;
		case GROUP_SUBSCRIBED:
			gtk_menu_set_active(GTK_MENU (menu), 1);
			break;
		case GROUP_NEW:
			gtk_menu_set_active(GTK_MENU (menu), 2);
			break;
		case GROUP_FOLDERS:
			gtk_menu_set_active (GTK_MENU (menu), 3);
			break;
		default: break;
	}
//        gtk_menu_set_active (GTK_MENU(menu), menu_item_index);
        gtk_option_menu_set_menu (GTK_OPTION_MENU(grouplist_mode_menu), menu);
        gtk_object_unref (GTK_OBJECT(menu));
        pan_unlock ();
}

typedef struct {
	gchar* name;
	int mode;
} GrouplistModeMenuStruct;

static GtkWidget*
grouplist_mode_menu_create (void)
{
	GtkWidget *option_menu = gtk_option_menu_new();
	GtkWidget *menu = gtk_menu_new ();
	int index = 0;
	int i;
	GrouplistModeMenuStruct foo[] =
	{
		{ _("All Groups"), GROUP_ALL },
		{ _("Subscribed"), GROUP_SUBSCRIBED },
		{ _("New Groups"), GROUP_NEW }
#if 0
		{ _("User Folders"), GROUP_FOLDERS }
#endif
	};
	const int row_qty = sizeof(foo) / sizeof(foo[0]);

	for (i=0; i<row_qty; ++i) {
		GtkWidget *item = gtk_menu_item_new_with_label (foo[i].name);
		gtk_signal_connect (GTK_OBJECT(item), "activate",
				    (GtkSignalFunc)grouplist_set_view_mode,
				    GINT_TO_POINTER(foo[i].mode));
		gtk_menu_append (GTK_MENU (menu), item);
		if (group_mode == foo[i].mode)
			index = i;
	}

	gtk_menu_set_active (GTK_MENU(menu), index);
        gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);
	gtk_widget_show_all (GTK_WIDGET(option_menu));

	return option_menu;
}

/*---[ grouplist_create ]---------------------------------------------
 *
 *--------------------------------------------------------------------*/
GtkWidget *
grouplist_create (void)
{
	GtkWidget *w;
        GtkWidget *vbox;
        GtkWidget *hbox;
	GtkWidget *toolbar = gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_TEXT);

	char *titles[] = {
		"",
//		"",
		_("Groups"),
		_("Unread"),
		_("Total"),
		_("Description")
	};

	vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_container_set_border_width (GTK_CONTAINER(vbox), GNOME_PAD_SMALL); 
	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);

	group_mode = gnome_config_get_int ("/Pan/State/group_mode=0");
	
	Pan.group_clist = gtk_clist_new_with_titles (5, titles);
	widget_set_font (GTK_WIDGET(Pan.group_clist), grouplist_font);
	gtk_clist_set_column_justification (GTK_CLIST(Pan.group_clist), 2, GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification (GTK_CLIST(Pan.group_clist), 3, GTK_JUSTIFY_RIGHT);


	w = gtk_label_new (_("Groups"));
	gtk_toolbar_append_widget(GTK_TOOLBAR(toolbar), w, "", "");
	gtk_toolbar_append_space (GTK_TOOLBAR(toolbar));

	grouplist_mode_menu = grouplist_mode_menu_create();
	gtk_widget_set (grouplist_mode_menu, "width", 120, "height", 25, NULL); /* FIXME: If we don't do this the om doesn't show up*/
	gtk_box_pack_start (GTK_BOX(hbox), grouplist_mode_menu, FALSE, FALSE, 0);

	w = gtk_label_new (_("Find:"));
	gtk_box_pack_start (GTK_BOX(hbox), w, FALSE, FALSE, 0);

	w = gtk_entry_new ();
	gtk_signal_connect (GTK_OBJECT (w), "activate",
			    GTK_SIGNAL_FUNC(grouplist_grep), w);
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), w,
		_("Type in a group search string and press ENTER.  "
		  "Wildcards are allowed."), "");
	gtk_toolbar_set_space_size(GTK_TOOLBAR(toolbar), 100);
	gtk_widget_show_all (toolbar);
	gtk_box_pack_start (GTK_BOX(hbox), w, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	w = gtk_scrolled_window_new (NULL, NULL);

	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (w),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER(w), Pan.group_clist);
        gtk_box_pack_start (GTK_BOX (vbox), w, TRUE, TRUE, 0);

	gtk_signal_connect (GTK_OBJECT (Pan.group_clist),
			    "select_row",
			    GTK_SIGNAL_FUNC (grouplist_select_row), NULL);

	gtk_signal_connect (GTK_OBJECT (Pan.group_clist),
			    "unselect_row",
			    GTK_SIGNAL_FUNC (grouplist_select_row), NULL);

	gtk_signal_connect (GTK_OBJECT (Pan.group_clist),
			    "button_press_event",
			    GTK_SIGNAL_FUNC (grouplist_button_press), NULL);

	gtk_clist_set_column_width     (GTK_CLIST (Pan.group_clist), 0, 16);
	gtk_clist_set_column_min_width (GTK_CLIST (Pan.group_clist), 0, 16);
	gtk_clist_set_column_max_width (GTK_CLIST (Pan.group_clist), 0, 16);

//	gtk_clist_set_column_width     (GTK_CLIST (Pan.group_clist), 1, 16);
//	gtk_clist_set_column_min_width (GTK_CLIST (Pan.group_clist), 1, 16);
//	gtk_clist_set_column_max_width (GTK_CLIST (Pan.group_clist), 1, 16);

	gtk_clist_set_column_width (GTK_CLIST (Pan.group_clist), 1, 170);
	gtk_clist_set_column_width (GTK_CLIST (Pan.group_clist), 2, 36);
	gtk_clist_set_column_width (GTK_CLIST (Pan.group_clist), 3, 33);
	gtk_clist_set_column_width (GTK_CLIST (Pan.group_clist), 4, 350);

//	gtk_clist_set_column_auto_resize (GTK_CLIST (group_clist), 1, TRUE);
//	gtk_clist_set_column_auto_resize (GTK_CLIST (group_clist), 2, TRUE);

/* JEL: this works well in paned mode, but not so well in notebook mode, we
   need to find a solution that will cover both ... maybe splitting the justification
   parts into the _construct funcs for both ...
	gtk_clist_set_column_justification ( GTK_CLIST(group_clist), 0, GTK_JUSTIFY_LEFT );
	gtk_clist_set_column_justification ( GTK_CLIST(group_clist), 1, GTK_JUSTIFY_RIGHT );
	gtk_clist_set_column_justification ( GTK_CLIST(group_clist), 2, GTK_JUSTIFY_RIGHT );
*/
	gtk_clist_set_selection_mode (GTK_CLIST (Pan.group_clist),
				      GTK_SELECTION_EXTENDED);

	group_mode = gnome_config_get_int ("/Pan/State/group_mode=0");
	if ( group_mode!=GROUP_ALL
	     && group_mode!=GROUP_SUBSCRIBED
	     && group_mode!=GROUP_NEW )
		group_mode = GROUP_ALL;

	/* Create the Popup Menu while we are here. */
	group_clist_menu = gnome_popup_menu_new (groups_menu_popup);	

	/* Lets have the Pixmap ready too */
//	disk_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d (disk_xpm);
	folder_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d (folder_xpm);
	group_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d (envelope_xpm);
//	subscribed_pixmap = (GnomePixmap *) gnome_pixmap_new_from_xpm_d (envelope_xpm);

	gtk_widget_show_all (vbox);
	return vbox;
}


/* Grep the group listing for whatever the user typed into the Find box */
static void
grouplist_grep (GtkWidget *group_entry)
{
	int i;
	Server *server = grouplist_get_current_server ();
	gchar *s, *pch;

	g_return_if_fail (Pan.group_clist);
	g_return_if_fail (server);

	pan_lock ();

	/* wrap the key in wildcards to make it a substring search...
	   unless there are already wildcards present, indicating the
	   user already knows what he wants */
        pch = gtk_editable_get_chars (GTK_EDITABLE(group_entry), 0, -1);
	pch = g_strstrip (pch);
	if (strchr(pch,'*'))
		s = g_strdup (pch);
	else if (*pch)
		s = g_strdup_printf("*%s*",pch);
	else
		s = NULL;
	g_free (pch);
	pch = NULL;

	gtk_clist_freeze (GTK_CLIST(Pan.group_clist));
	gtk_clist_clear (GTK_CLIST(Pan.group_clist));
	for (i=server->groups->len-1; i>=0; --i)
	{
		Group* group = (Group*) g_ptr_array_index (server->groups, i);

		if (group_mode!=GROUP_ALL && !(group->flags&group_mode))
			continue;
		if (s && *s && fnmatch(s,group->name,0))
			continue;

		grouplist_add_group (server, group);
	}
	gtk_clist_thaw (GTK_CLIST(Pan.group_clist));
	pan_unlock();

	g_free (s);
}


typedef struct
{
	Server * server;
	GPtrArray * groups;
	StatusItem * status;
}
LoadStruct;

static void
grouplist_load_group (const char *key, const char *val, gpointer data)
{
	LoadStruct * load_data = (LoadStruct*) data;
	Group * g = NULL;
	const char * pch;
	char * newsrc_ranges;

	/* create the group */
	g = group_new (load_data->server, key);
	g->flags             = get_next_token_int (val, '\n', &pch);
	g->description       = get_next_token_str (pch, '\n', &pch);
	g->state_filter      = get_next_token_int (pch, '\n', &pch);
	g->download_dir      = get_next_token_str (pch, '\n', &pch);
	g->sort_style        = get_next_token_int (pch, '\n', &pch);
	g->article_low       = get_next_token_ulong (pch, '\n', &pch);
	g->article_high      = get_next_token_ulong (pch, '\n', &pch);
	g->article_qty       = get_next_token_int (pch, '\n', &pch);
	g->article_read_qty  = get_next_token_int (pch, '\n', &pch);
	newsrc_ranges        = get_next_token_str (pch, '\n', &pch);
	newsrc_init (g->newsrc, newsrc_ranges, g->article_low, g->article_high);
	g_free (newsrc_ranges);

	pan_object_ref (PAN_OBJECT(g));
	pan_object_sink (PAN_OBJECT(g));
        g_ptr_array_add (load_data->groups, g);

	/* tell the status listeners where we are.
	   509 is just a prime, not too big, not to small
	   number we're using for the frequency of UI updates */
	if (!(load_data->groups->len % 509))
		status_item_emit_status (
			load_data->status,
			_("Loaded %d Groups"),
			load_data->groups->len);
}

static gchar*
grouplist_load_describe (const StatusItem* item)
{
	return g_strdup (_("Loading Groups"));
}

/* Load a grouplist database into memory (in a separate thread) */
static void
grouplist_load_thread (Server *server)
{
	gchar *path;
	pan_db db;
	StatusItem *item = NULL;
	LoadStruct load_data;

	/* create a new status item to get load messages */
	item = STATUS_ITEM(status_item_new(grouplist_load_describe));
	pan_object_ref(PAN_OBJECT(item));
	pan_object_sink(PAN_OBJECT(item));
	gui_add_status_item (item);

	/* open the database... */
	path = g_strdup_printf ("/%s/%s_grouplist.db", data_dir, server->name);
	db = pan_db_open (path);
	g_free (path);

	/* load the groups */
	if (db != NULL) {
		load_data.server = server;
		load_data.groups = server->groups;
		load_data.status = item;
		pan_db_foreach (db, grouplist_load_group, &load_data);
		pan_db_close (db);
	}

	/* clean out the status item -- grouplist_update_ui has its own
	   status item, so we don't need this one anymore. */
	gui_remove_status_item (item);
	pan_object_unref(PAN_OBJECT(item));
	item = NULL;

	/* update the group list */
	grouplist_update_ui_thread (server);
}

/**
 * grouplist_load:
 * @server: The server of the grouplist to load
 *
 * Load a grouplist database into memory
 **/
void
grouplist_load (Server *server)
{
	pthread_t thread;
	pthread_create (&thread, NULL, (void*)grouplist_load_thread, server);
	pthread_detach (thread);
}

static void
group_serialize (const Group * g, GString * setme)
{
	g_return_if_fail (g!=NULL);
	g_return_if_fail (setme!=NULL);

	g_string_sprintf (setme,
		"%d\n" "%s\n" "%u\n"
		"%s\n" "%d\n"
		"%lu\n" "%lu\n" "%d\n" "%d\n",

		g->flags,
		(g->description!=NULL ? g->description : ""),
		g->state_filter,

		(g->download_dir!=NULL ? g->download_dir : ""),
		g->sort_style,

		g->article_low,
		g->article_high,
		g->article_qty,
		g->article_read_qty);

	/* append the newsrc string */
	newsrc_get_read_str (g->newsrc, setme);
	g_string_append_c (setme, '\n');
}

void
grouplist_save_group (Group * group)
{
	GString * buf = g_string_sized_new (4096);
	pan_db db = 0;

	pan_warn_if_fail (group != NULL);
	pan_warn_if_fail (group->server != NULL);
	pan_warn_if_fail (is_nonempty_string (group->server->name));

	/* get the database */
	g_string_sprintf (buf, "/%s/%s_grouplist.db",
		data_dir,
		group->server->name);
	db = pan_db_open (buf->str);
	g_string_truncate (buf, 0);


	/* store the info */
	if (db != NULL)
	{
		group_serialize (group, buf);
		pan_db_put_value_str (db, group->name, buf->str);
		pan_db_close (db);
	}

	/* cleanup */
	g_string_free (buf, TRUE);
}


static void
grouplist_save_list (const Server * server,
                     Group ** groups,
		     int qty)
{
	GString * buf = g_string_sized_new (4096);
	pan_db db;

	g_string_sprintf (buf, "/%s/%s_grouplist.db", data_dir, server->name);
	db = pan_db_open (buf->str);
	
	if (db != NULL)
	{
		int i;
		for (i=0; i<qty; ++i)
		{
			Group * g = groups[i];
			group_serialize (g, buf);
			pan_db_put_value_str (db, g->name, buf->str);
			g->dirty = 0;
		}

		pan_db_close (db);
	}

	g_string_free (buf, TRUE);
}

void
grouplist_save_selected (const Server * server)
{
	GtkCList *c;
	GPtrArray *groups;
	GList *l;

	/* build the array of groups */
       	c = GTK_CLIST(Pan.group_clist);
       	groups = g_ptr_array_new();
	for (l=c->selection; l; l=l->next)
		g_ptr_array_add (groups, gtk_clist_get_row_data (c, GPOINTER_TO_INT(l->data)));

	/* save the groups */
	grouplist_save_list (server, (Group**)groups->pdata, groups->len);

	/* cleanup */
	g_ptr_array_free (groups, TRUE);
}

void
grouplist_save_dirty (Server * server)
{
	int i;
	GPtrArray * a;

	/* entry assertions */
	g_return_if_fail (server != NULL);
	g_return_if_fail (server->groups != NULL);

	/* build the list of dirty groups */
	a = g_ptr_array_new ();
	for (i=0; i!=server->groups->len; ++i) {
		Group * g = GROUP (g_ptr_array_index(server->groups, i));
		if (g->dirty)
			g_ptr_array_add (a, g);
	}

	/* save the dirty groups */
	if (a->len != 0) {
		printf (_("Pan is saving changed newsgroups for %s... "),
			server->name);
		fflush (NULL);
		grouplist_save_list (server, (Group**)a->pdata, a->len);
		printf (_("%d saved.\n"), a->len);
	}

	/* cleanup */
	g_ptr_array_free (a, TRUE);
}

/**
***
**/

void
grouplist_set_selected_groups (GSList * sel)
{
	GSList * l;
	GtkCList * clist = GTK_CLIST(Pan.group_clist);

	pan_lock ();
	gtk_clist_freeze (clist);
	gtk_clist_unselect_all (clist);
	for (l=sel; l!=NULL; l=l->next)
	{
		const int row = gtk_clist_find_row_from_data (clist, GROUP(l->data));
		gtk_clist_select_row (clist, row, 0);
		if (!gtk_clist_row_is_visible (clist, row))
		{
			gfloat valign = 0.5;
			gfloat halign = 0.0;
			gtk_clist_moveto (clist, row, 0, (double)valign, (double)halign);
		}
	}
	gtk_clist_thaw (clist);
	pan_unlock ();
}

/**
***
**/

/*---[ grouplist_update ]---------------------------------------------
 * update the entire group CLIST, clearing and then filling it in
 * with the groups that are part of the current group mode (all, 
 * subscribed, etc.)
 *--------------------------------------------------------------------*/

static gchar*
grouplist_update_describe (const StatusItem *item)
{
	return g_strdup (_("Updating Group List"));
}

static void
grouplist_update_ui_thread (void* data)
{
	Server * server = (Server*)data;
	GtkCList * clist = GTK_CLIST(Pan.group_clist);
	GSList * sel;
	StatusItem *item = NULL;
	int i;

	g_return_if_fail (server != NULL);

        /* create a new status item to get sort/thread messages */
	item = STATUS_ITEM(status_item_new(grouplist_update_describe));
	pan_object_ref(PAN_OBJECT(item));
	pan_object_sink(PAN_OBJECT(item));
	gui_add_status_item (item);
	status_item_emit_init_steps (item, server->groups->len);
	status_item_emit_status (item, _("Updating Group List"));

	pan_lock();

	/* set the groups column title */
	if (group_mode == GROUP_SUBSCRIBED)
		gtk_clist_set_column_title (clist, 1, _("Subscribed Groups"));
	else if (group_mode == GROUP_ALL)
		gtk_clist_set_column_title (clist, 1, _("All Groups"));
	else if (group_mode == GROUP_NEW)
		gtk_clist_set_column_title (clist, 1, _("New Groups"));
	else if (group_mode == GROUP_FOLDERS)
		gtk_clist_set_column_title (clist, 1, _("Folders"));

	/* clear the clist */
	sel = grouplist_get_selected_groups ();
	gtk_clist_freeze (clist);
	gtk_clist_clear (clist);


	/* add the groups */
	for (i=server->groups->len-1; i>=0; --i)
	{
		Group* group = (Group*) g_ptr_array_index (server->groups, i);
		if ((group_mode==GROUP_ALL) || (group_mode & group->flags))
		{
			grouplist_add_group (server, group);
			pan_unlock();
			status_item_emit_next_step (item);
			pan_lock();
		}
	}
	gtk_clist_thaw (clist);
	pan_unlock();

	/* update the selection */
	grouplist_set_selected_groups (sel);
	g_slist_free (sel);

	/* clean out the status item */
	gui_remove_status_item (item);
	pan_object_unref(PAN_OBJECT(item));
}

void
grouplist_update_ui (Server *server)
{
	pthread_t thread;

	g_return_if_fail (server != NULL);
	pthread_create (&thread, NULL, (void*)grouplist_update_ui_thread, server);
	pthread_detach (thread);
}


#if 0
static void
grouplist_add_folder (Group *group)
{
	int row;
	char total_buf[16] = { '\0' };
	char unread_buf[16] = { '\0' };
	char * const text[7] = {
		"",
//		"",
		group->name,
		unread_buf,
		total_buf,
		(group->description?group->description:""),
		NULL
	};
	const gint total = group->article_qty;
	if (group->article_low != 0)
	{
		const gint unread = MAX(0, total - group->article_read_qty);
		commatize_ulong (total, total_buf);
		commatize_ulong (unread, unread_buf);
	}

	row = gtk_clist_prepend (GTK_CLIST(Pan.group_clist), (char**)text);

	gtk_clist_set_pixmap (
		GTK_CLIST(Pan.group_clist), row, 0,
		folder_pixmap->pixmap,
		folder_pixmap->mask);
/*
	if (total)
		gtk_clist_set_pixmap (
			GTK_CLIST(Pan.group_clist), row, 1,
			disk_pixmap->pixmap,
			disk_pixmap->mask);
*/

	gtk_clist_set_row_data (GTK_CLIST(Pan.group_clist), 0, group);
}
#endif

/* note that this is called inside a pan_lock() */
static void
grouplist_add_group (const Server *server, Group *group)
{
	int row;
	char total_buf[16] = { '\0' };
	char unread_buf[16] = { '\0' };
	char * const text[7] = {
		"",
		group->name,
		unread_buf,
		total_buf,
		(group->description?group->description:""),
		NULL
	};

	const gint total = group->article_qty;
	const gint unread = MAX(0, total - group->article_read_qty);
	commatize_ulong (total, total_buf);
	commatize_ulong (unread, unread_buf);

	row = gtk_clist_prepend (GTK_CLIST(Pan.group_clist), (char**)text);

	if (group_is_subscribed (group))
		gtk_clist_set_pixmap (
			GTK_CLIST(Pan.group_clist), row, 0,
			group_pixmap->pixmap,
			group_pixmap->mask);

	gtk_clist_set_row_data (GTK_CLIST(Pan.group_clist), 0, group);
}

/* Remove a single row from the group CList, useful in unsubscribing */
void
grouplist_remove_row (const Group *group)
{
	GtkCList *list = GTK_CLIST(Pan.group_clist);

	pan_lock();
	gtk_clist_remove (list, gtk_clist_find_row_from_data (list,(gpointer)group));
	pan_unlock();
}

/**
 * grouplist_update_row:
 * @group: Group who's row is to be updated.
 *
 * Update the CList row for a group to show it's message counts or
 * any other relative data (subscribed/on disk) for that group.
 **/
void
grouplist_update_row (const Group *group)
{
	char buf[16];
	gint row = 0;
	const gint total = group->article_qty;
	const gint unread = MAX(0, total-group->article_read_qty);
	GtkCList * list = GTK_CLIST (Pan.group_clist);

	pan_lock();

	row = gtk_clist_find_row_from_data (list,(gpointer)group);

	if (group_is_subscribed (group))
		gtk_clist_set_pixmap (
			GTK_CLIST(Pan.group_clist), row, 0,
			group_pixmap->pixmap,
			group_pixmap->mask);
	else if (group->type != TYPE_USER_FOLDER)
		gtk_clist_set_text (GTK_CLIST(Pan.group_clist),
				    row, 0, "");

	/* total column */
	commatize_ulong (total, buf);
	gtk_clist_set_text (list, row, 3, buf);

	/* unread column */
	commatize_ulong (unread, buf);
	gtk_clist_set_text (list, row, 2, buf);

	pan_unlock();
}

static void
grouplist_update_selected_rows (void)
{
	GtkCList *list = GTK_CLIST(Pan.group_clist);
	GList *l = list->selection;
	for (l=list->selection; l!=NULL; l=l->next)
	{
		Group *group = NULL;
		int index = GPOINTER_TO_INT(l->data);
		group = (Group*) gtk_clist_get_row_data (list, index);
		grouplist_update_row (group);
	}
}

static void
grouplist_internal_select_row (int row)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);

	pan_lock();
	gtk_clist_unselect_all (list);
	gtk_clist_select_row (list,row,-1);
	pan_unlock();
}

void
grouplist_prev_group (void)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);
        const int row_qty = list->rows;
       	int row = 0;

	if (!row_qty)
		return;

	if ( list->selection ) {
		row = GPOINTER_TO_INT(list->selection->data) - 1;
		if (row<0)
			row = row_qty-1;
	}

	grouplist_internal_select_row (row);
	articlelist_set_current_group (GROUP(gtk_clist_get_row_data(list,row)));
}

void
grouplist_next_group (void)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);
        const int row_qty = list->rows;
       	int row = 0;

	if (!row_qty)
		return;

	if ( list->selection )
		row = (GPOINTER_TO_INT(list->selection->data)+1) % row_qty;

	grouplist_internal_select_row (row);
	articlelist_set_current_group (GROUP(gtk_clist_get_row_data(list,row)));
}

static gboolean
grouplist_select_row_if_unread (int row)
{
        GtkCList * list = GTK_CLIST(Pan.group_clist);
	Group * g = gtk_clist_get_row_data(list,row);
	const gulong total = g->article_qty;
	const gulong read = g->article_read_qty;
	gboolean retval = FALSE;

	if (read < total)
	{
		grouplist_internal_select_row (row);
		articlelist_set_current_group (g);
		retval = TRUE;
	}

	return retval;
}

void
grouplist_next_unread_group (void)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);
        const int row_qty = list->rows;
	int i;
       	int row = 0;

	if (!row_qty)
		return;

	/* get the first row */
	if ( list->selection )
		row = (GPOINTER_TO_INT(list->selection->data) + 1) % row_qty;
	else
		row = 0;

	/* walk through */
	for ( i=0; i<row_qty; ++i, row=++row%row_qty )
		if ( grouplist_select_row_if_unread (row) )
			return;
}

void
grouplist_prev_unread_group (void)
{
        GtkCList* list = GTK_CLIST(Pan.group_clist);
        const int row_qty = list->rows;
       	int row = 0;
	int i;

	if (!row_qty)
		return;

	/* get the first row */
	if ( list->selection ) {
		row = GPOINTER_TO_INT(list->selection->data);
		row = row ? row-1 : row_qty-1;
	}
	else
		row = row_qty-1;

	/* walk through */
	for ( i=0; i!=row_qty; ++i, row=row?--row:row_qty-1 )
		if ( grouplist_select_row_if_unread (row) )
			return;
}
